unit LoadFlow;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, StdCtrls, Skemadef, Utildef, Buttons,
  Nodes,Elplex,Compdef, Imagodef, Skemproc, Menus, Cmatmul,GenVolt,Matmul;

procedure LoadFlowCalculation(ipSet : longint; probl: Solve; var success : boolean); {solve defined in elplex.pas}
procedure unconnected(pSet: TpCompSet);

type
{class for performing the load flow calculations}
  TpLoadFlow = ^TLoadFlow;         {define pointer type first !}
  TLoadFlow = class(TSetVoltage)   {derive the class}
public
  constructor Create(ipSet : longint; var success : boolean);
  destructor  Destroy; override;
  procedure   ClearInitially;
  procedure   FindLoadsAndSlackBus;
  procedure   FindSlackBus;
  procedure   GetMemoryForOtherDerivates;
  procedure   FreeMemoryForOtherDerivates;
  procedure   CalculateOtherGeneratorDerivatives;
  procedure   CalculateLoadDerivatives;
  procedure   SolveLoadFlow(probl: Solve);
private
  nDrain_lf  : integer;     {number of current drains (loads)}
  nSlack_lf  : integer;     {number of slack nodes (sources)}
  pdLUdGU_lf : pComplex;    {pointer to d(load voltage)/d(gener.voltage)}
  pdGUdLI_lf : pComplex;    {pointer tp d(gen.voltage )/d(load current)}
  pdGIdLI_lf : pComplex;    {pointer to d(generator current)/d(load current)}
  pdLUdLI_lf : pComplex;    {pointer to d(load voltage)/d(load current)}
  pSolut_lf  : pComplex;    {pointer to latest voltages, currents in the net}
  pSource_lf : pComplex;    {rhs = source voltages and currents}
  ppDrain_lf : TppComp;     {pointer to an array of drain pointers}
  pSet_lf    : TpCompSet;   {pointer to the common}
  pSlack_lf  : TpComp;      {pointer to the slack bus}
end;

var
  LoadFlower : TLoadFlow;   {object for performing the load flow calculations}

implementation

uses
  imagopas;

constructor TLoadFlow.Create(ipSet : longint; var success : boolean);
{reserve space, initialize
 ipSet   pointer to the set of components etc, as integer}
var
  cont1,cont2  : boolean;  {continue with load flow}
  startSuccess : boolean;  {first part of create was successful}
begin
  ClearInitially;                  {clear all variables initially}
  inherited create(ipSet,startSuccess);
  pSet_lf := TpCompSet(ipSet);     {pointer to the common}
  nDrain_lf := n1_gv+n3_gv;        {copy number of drains + loads from temp.}
  nSlack_lf := n2_gv;              {salck bus = feeder}
  {continue, if there are generators or loads and feeder(s)}
  if (startSuccess) then begin     {there are generators}
    {check desired powers of generators}
    if (IsDesiredPowerPositive) then begin  {positive power, can continue}
      cont1 := true;
    end else begin                          {not positive power, cannot continue}
      cont1 := false;
      show_message('nonpositive desired power in some generator(s)');
    end;
  end else begin
    cont1 := false;                 {no generators}
  end;
  {check loads and feeders}
  cont2 := ((nDrain_lf > 0) and (nSlack_lf > 0));
  if (cont1 or cont2) then begin            {generators, or loads+feeder(s)}
    getmem(pSolut_lf ,  nnod_gv*SizeOf(complex)); {for latest volt,curr.}
    getmem(pSource_lf,  nnod_gv*SizeOf(complex)); {for latest volt,curr.}
    if (nSlack_lf > 0) then begin  {there is a slack bus }
      if (nDrain_lf > 0) then begin
        {get memory for pointers to drains}
        getmem(ppDrain_lf,nDrain_lf*SizeOf(TpComp));
        {get memory for the derivatives of generator currents and load
         voltages with respect to generator voltages and load currents}
        GetMemoryForOtherDerivates;
        FindLoadsAndSlackBus;        {save pointer to slack bus, + loads}
      end else begin
        FindSlackBus;
      end;
      success := true;
    end else begin
      pSlack_lf := nil;
      success := false;
      show_message('no slack bus, thus no calculation');
    end;
  end else begin      {no generators or no loads+feeder(s)}
    show_message('No generator or no load, thus no calculation');
    {set flag}
    success := false;
  end;
end;

destructor TLoadFlow.Destroy;
{reserve space, initialize
 ipSet   pointer to the set of components etc, as integer}
begin
  if (pSolut_lf  <> nil) then freemem(pSolut_lf  ,  nnod_gv*SizeOf(complex));
  if (pSource_lf <> nil) then freemem(pSource_lf ,  nnod_gv*SizeOf(complex));
  if (ppDrain_lf <> nil) then freemem(ppDrain_lf , nDrain_lf*SizeOf(TpComp));
  FreeMemoryForOtherDerivates;
  inherited destroy;
end;

procedure TLoadFlow.ClearInitially;
{clear all variables initially}
begin
  nDrain_lf  := 0;
  nSlack_lf  := 0;
  pdLUdGU_lf := nil;
  pdGUdLI_lf := nil;
  pdGIdLI_lf := nil;
  pdLUdLI_lf := nil;
  pSolut_lf  := nil;
  pSource_lf := nil;
  ppDrain_lf := nil;
  pSlack_lf  := nil;
end;

procedure TLoadFlow.FindLoadsAndSlackBus;
{find the pointers to loads, put them in the arrays
 for respective pointers
 find also the (a) slack bus = feeder}
var
  errc    : integer;  {error flag, indicates end of components}
  jcomp   : integer;  {counts components}
  ModNam  : string;   {name of a component}
  pComp   : TpComp;   {pointer to a component}
  ppDrain : TppComp;  {pointer to load (drain) pointer}
begin
  {scan all components in order to find loads (drains) and generators and
   the slack bus}
  errc := 0;
  pComp := nil;
  jcomp := 0;
  pSlack_lf := nil;
  ppDrain := ppDrain_lf;
  while (errc = 0) do begin                {scan all components}
    pSet^.NextComp(pComp,errc,jcomp,true); {get pointer to next component}
    if (errc = 0) then begin               {component found}
      pComp^.GetModName(ModNam);           {get the name (type) of the component}
      if (ModNam = gCurr) or (ModNam = gLoad) then begin  {is it a drain (load)}
        ppDrain^:= pComp;
        inc(ppDrain);
      end else if (ModNam = gFeed) then begin {feeder ?}
        pSlack_lf := pComp;            {save pointer to (latest) slack bus}
      end;
    end;
  end;
end;

procedure TLoadFlow.FindSlackBus;
{find the pointer the (a) slack bus = feeder}
var
  errc    : integer;  {error flag, indicates end of components}
  jcomp   : integer;  {counts components}
  ModNam  : string;   {name of a component}
  pComp   : TpComp;   {pointer to a component}
begin
  {scan all components in order to find loads (drains) and generators and
   the slack bus}
  errc := 0;
  pComp := nil;
  jcomp := 0;
  pSlack_lf := nil;
  while (errc = 0) do begin                {scan all components}
    pSet^.NextComp(pComp,errc,jcomp,true); {get pointer to next component}
    if (errc = 0) then begin               {component found}
      pComp^.GetModName(ModNam);           {get the name (type) of the component}
      if (ModNam = gFeed) then begin       {feeder ? = slack bus}
        pSlack_lf := pComp;                {save pointer to (first) slack bus}
        break;
      end;
    end;
  end;
end;

procedure TLoadFlow.GetMemoryForOtherDerivates;
{get memory for the derivatives of generator currents and load
 voltages with respect to generator voltages and load currents}
begin
  {check, just to be sure}
  if (nGener_gv > 0) and (nDrain_lf > 0) then begin
    getmem(pdLUdGU_lf,nDrain_lf*nGener_gv*SizeOf(complex));
    getmem(pdGUdLI_lf,nGener_gv*nDrain_lf*SizeOf(complex));
    getmem(pdGIdLI_lf,nGener_gv*nDrain_lf*SizeOf(complex));
  end else begin
    pdLUdGU_lf := nil;
    pdGUdLI_lf := nil;
    pdGIdLI_lf := nil;
  end;
  if (nDrain_lf > 0) then begin
    getmem(pdLUdLI_lf,nDrain_lf*nDrain_lf*SizeOf(complex));
  end else begin
    pdLUdLI_lf := nil;
  end;
end;

procedure TLoadFlow.FreeMemoryForOtherDerivates;
{get memory for the derivatives of generator currents and load
 voltages with respect to generator voltages and load currents}
begin
  {check, just to be sure}
  if (nGener_gv > 0) and (nDrain_lf > 0) then begin
    freemem(pdLUdGU_lf,nDrain_lf*nGener_gv*SizeOf(complex));
    freemem(pdGUdLI_lf,nGener_gv*nDrain_lf*SizeOf(complex));
    freemem(pdGIdLI_lf,nGener_gv*nDrain_lf*SizeOf(complex));
  end;
  if (nDrain_lf > 0) then begin
    freemem(pdLUdLI_lf,nDrain_lf*nDrain_lf*SizeOf(complex));
  end;
end;

procedure TLoadFlow.CalculateOtherGeneratorDerivatives;
{calculate the derivatives of generator current and load voltage
 with respect to generator voltage .
 When entering, the set of equations has been decomposed, but not
 yet backsubstituted}
var
  cpu      : complex;  {temporary}
  err      : integer;  {error flag}
  ii,jj    : integer;  {indices}
  pGen     : TpComp;   {pointer to a generator}
  ppComp   : TppComp;  {pointer to component pointers}
  ppGener  : TppComp;  {pointer to an array of pointers}
  VoltEq   : integer;  {number of the voltage equation = current "node"}
begin
  {calculate the derivatives}
  ppGener := ppGener_gv;                        {p. to array of gen.pointers}
  for ii := 1 to nGener_gv do begin             {scan all generators}
    pGen := ppGener^;                           {pointer to this generator}
    inc(ppGener);
    pGen^.GetNode(2,VoltEq,err);                {get voltage equation number}
    {for derivatives, just put one to the rhs}
    cpu.x := 0;
    cpu.y := 0;                                 {clear first}
    for jj:=1 to nnod_gv do putCvec(pWs,jj,nnod_gv,cpu);
    cpu.x := 1.0;
    cpu.y := 0.0;
    putCvec(pWs,VoltEq,nnod_gv,cpu);            {put 1 to the "deriver"}
    CLUbacksub(pAdmat,nnod_gv,pIndex,pWs);      {get "solution" }
    {find change in current for all generators}
    ppComp := ppGener_gv;                       {pointer to first generator}
    for jj:=1 to nGener_gv do begin
      PickDerivative(ppComp,2,cpu);
      putCmat(pdGIdGU_gv,jj,ii,nGener_gv,nGener_gv,cpu); {save it}
    end;
    if (nDrain_lf > 0) then begin               {only if there are loads}
      {find change in voltage for all loads}
      ppComp := ppDrain_lf;                       {pointer to first load}
      for jj:=1 to nDrain_lf do begin
        PickDerivative(ppComp,1,cpu);
        putCmat(pdLUdGU_lf,jj,ii,nDrain_lf,nGener_gv,cpu); {save it}
      end;
    end;
  end;
//showCmat1(pdgidgu_gv,nGener_gv,nGener_gv);
end;

procedure TLoadFlow.CalculateLoadDerivatives;
{calculate the derivatives of generator current and and voltage,
 and load voltage with respect to load current.
 When entering, the set of equations has been decomposed, but not
 yet backsubstituted}
var
  cpu      : complex;  {temporary}
  CurrEq   : integer;  {number of current equation = voltage node}
  err      : integer;  {error flag}
  ii,jj    : integer;  {indices}
ix,iy : integer;
lbl   : string;
  pLoad    : TpComp;   {pointer to a generator}
  ppComp   : TppComp;  {pointer to component pointers}
  ppLoad   : TppComp;  {pointer to an array of pointers}
begin
  if (nDrain_lf > 0) then begin                 {only, if there are loads}
    {calculate the derivatives}
    ppLoad := ppDrain_lf;                         {p. to array of load pointers}
    for ii := 1 to nDrain_lf do begin             {scan all loads}
      pLoad := ppLoad^;                           {pointer to this generator}
      inc(ppLoad);
      pLoad^.GetNode(2,CurrEq,err);               {get current equation number}
      {clear first}
      cpu.x := 0;
      cpu.y := 0;                                 {clear first}
      for jj:=1 to nnod_gv do putCvec(pWs,jj,nnod_gv,cpu);
      {for derivatives, just put one to the rhs}
      cpu.x := 1.0;
      cpu.y := 0.0;
      putCvec(pWs,CurrEq,nnod_gv,cpu);            {put 1 to the "deriver"}
      CLUbacksub(pAdmat,nnod_gv,pIndex,pWs);      {get "solution" }
      if (nGener_gv > 0 ) then begin
        {find change in voltage  for all generators}
        ppComp := ppGener_gv;                       {pointer to first generator}
        for jj:=1 to nGener_gv do begin
          PickDerivative(ppComp,1,cpu);             {in GenVolt.pas}
          putCmat(pdGUdLI_lf,jj,ii,nGener_gv,nDrain_lf,cpu); {save it}
        end;
        {find change in current for all generators}
        ppComp := ppGener_gv;                       {pointer to first generator}
        for jj:=1 to nGener_gv do begin
          PickDerivative(ppComp,2,cpu);             {in GenVolt.pas}
          putCmat(pdGIdLI_lf,jj,ii,nGener_gv,nDrain_lf,cpu); {save it}
        end;
      end;
      {find change in voltage for all loads}
      ppComp := ppDrain_lf;                       {pointer to first load}
      for jj:=1 to nDrain_lf do begin
//ppComp^.GetLabel(lbl,ix,iy);
        PickDerivative(ppComp,1,cpu);               {in GenVolt.pas}
        putCmat(pdLUdLI_lf,jj,ii,nDrain_lf,nDrain_lf,cpu); {save it}
//if (ii = jj) then showmessage(lbl+'  '+inttostr(ii)+'  '+floattostr(cabs(cpu)));
      end;
    end;
//showCmat1(pdLUdLI_lf,nDrain_lf,nDrain_lf);
  end;
end;

procedure ReserveForEquations(nn : integer; var paa,pbb,pww : pDouble;
                                            var pind : pInteger);
{reserve and clear memory for a set of linear equations
 nn    number of equations
 paa   pointer to the equations
 pbb   pointer to right hand side
 pww   pointer to working space
 pind  pointer to indcies for row excnages}
var
  ii,jj  : integer;  {indices}
begin
  if (nn > 0) then begin
    getmem(paa,nn*nn*SizeOf(double));  {equations}
    getmem(pbb,nn*SizeOf(double));
    getmem(pww,nn*SizeOf(double));
    getmem(pind,nn*SizeOf(integer));
    {clear initially}
    for ii:=1 to nn do begin
      for jj:=1 to nn do begin
        putDmat(paa,ii,jj,nn,nn,0.0);
      end;
      putDvec(pbb,ii,nn,0.0);
      putDvec(pww,ii,nn,0.0);
      putIvec(pind,ii,nn,0);
    end;
  end else begin
    paa := nil;
    pbb := nil;
    pww := nil;
    pind := nil;
  end;
end;

procedure TLoadFlow.SolveLoadFlow(probl: Solve);
{solve the load flow problem}
const
  niter = 25;
var
  apu,apv : single;   {temporary}
  apw     : single;   {temporary}
  ccrit   : single;   {convergence crit.}
  cfgoal  : single;   {desired load cos(fii)}
  converged:boolean;  {true, if converged}
  CorrSum : single;   {sum of (abs) correction}
  cpu,cpv : complex;  {temporary}
  dpu,dpv : double;   {temporary}
  ii,jj,kk: integer;  {indices}
  dIrdEr  : single;   {d(gener.real current)/d(gener.real internal voltage)}
  dIrdEi  : single;   {d(gener.real current)/d(gener.imag.internal voltage)}
  dUrdEr  : single;   {d(gener.real voltage)/d(gener.real internal voltage)}
  dUrdEi  : single;   {d(gener.real voltage)/d(gener.imag.internal voltage)}
  dIidEr  : single;   {d(gener.imag.current)/d(gener.real internal voltage)}
  dIidEi  : single;   {d(gener.imag.current)/d(gener.imag.internal voltage)}
  dUidEr  : single;   {d(gener.imag.voltage)/d(gener.real internal voltage)}
  dUidEi  : single;   {d(gener.imag.voltage)/d(gener.imag.internal voltage)}
  dIrdJr  : single;   {d(gener.real current)/d(real load current)}
  dIrdJi  : single;   {d(gener.real current)/d(imag.load current)}
  dUrdJr  : single;   {d(gener.real voltage)/d(real load current)}
  dUrdJi  : single;   {d(gener.real voltage)/d(imag.load current)}
  dIidJr  : single;   {d(gener.imag.current)/d(real load current)}
  dIidJi  : single;   {d(gener.imag.current)/d(imag.load current)}
  dUidJr  : single;   {d(gener.imag.voltage)/d(real load current)}
  dUidJi  : single;   {d(gener.imag.voltage)/d(imag.load current)}
  dVrdEr  : single;   {d(load real voltage)/d(gener.real internal voltage)}
  dVrdEi  : single;   {d(load real voltage)/d(gener.imag.internal voltage)}
  dVidEr  : single;   {d(load imag.voltage)/d(gener.real internal voltage)}
  dVidEi  : single;   {d(load imag.voltage)/d(gener.imag.internal voltage)}
  dVrdJr  : single;   {d(load real voltage)/d(real load current)}
  dVrdJi  : single;   {d(load real voltage)/d(imag.load current)}
  dVidJr  : single;   {d(load imag.voltage)/d(real load current)}
  dVidJi  : single;   {d(load imag.voltage)/d(imag.load current)}
  dGIdGU  : complex;  {d(gener.current)/d(gener.voltage)}
  dGIdGUii: complex;  {d(gener.current)/d(gener.voltage) for generator ii}
  dGUdGU  : complex;  {d(gener.output voltage)/d(internal gener.voltage)}
  dGIdLI  : complex;  {d(gener.current)/d(load current)}
  dGUdLI  : complex;  {d(gener.output voltage)/(d(load current)}
  dLUdGU  : complex;  {d(load voltage)/d(gener.voltage)}
  dLUdLI  : complex;  {d(load voltage)/d(load current)}
  Dodd    : single;   {flag for row exchanges}
  Egen1   : complex;  {internal voltage of a generator}
  Egen2   : complex;  {internal voltage of a generator}
  Igen    : complex;  {generator current}
  Iload   : complex;  {load current}
  iparal  : integer;  {index of (eventual) parallel generator}
  iter    : integer;  {iteration counter}
  ix,iy   : integer;  {position of the label}
  lbl     : string;   {label of the component}
  line    : string;   {for tests}
  ModNam  : string;   {name of component}
  ngp     : single;   {number of parallel generators}
  nn      : integer;  {total number of unknowns}
  paa     : pDouble;  {pointer to the matrix for the equations}
  pbb     : pDouble;  {pointer to the right hand side}
  pdd     : pDouble;  {temporary pointer}
  pind    : pInteger; {pointer to indices, for row exchanges}
  pComp   : TpComp;   {pointer to a component}
  pGener  : TpComp;   {pointer to a generator}
  pLoad   : TpComp;   {pointer to a load (a drain)}
  p2Gen   : TpComp;   {pointer to another generator}
  pww     : pDouble;  {pointer to working space}
  ppGener : TppComp;  {pointer to the array of generator pointers}
  ppLoad  : TppComp;  {pointer to the array of load pointers}
  PGgoal  : single;   {desired generator power}
  PLgoal  : single;   {desired load real power}
  QLgoal  : single;   {desired load imag.power}
  Sngl    : integer;  {index of row exhanges}
  Tout    : double;   {test output}
  Ugen    : complex;  {generator voltage}
  Uload   : complex;  {load current}
  U2goal  : single;   {desired generator voltage squared}
  VAgoal  : single;   {desired load (VA) power}
begin
  {reserve space for the equations}
  if (nGener_gv > 0) or ( nDrain_lf > 0) then begin  {there are generators or loads}
    nn := 2*(nGener_gv+nDrain_lf);          {2, because real and imag.part}
    ReserveForEquations(nn,paa,pbb,pww,pind);
    {copy initial rhs = sources}
    CopyCvec(pRhs,pSource_lf,nnod_gv);
    {copy initial (nominal) solution}
    CopyCvec(pNomSol_gv,pSolut_lf,nnod_gv);
    {find largest source, for converg.crit.}
    apu := 0;
    for ii:=1 to nnod_gv do begin
      getCvec(pRhs,ii,nnod_gv,cpu);
      apv := Cabs(cpu);
      if (apv > apu) then apu := apv;
    end;
    if (apu > 1.0) then ccrit := 1.e-5*apu
    else                ccrit := 1.e-5;
    {iterate a few times}
    converged := false;

//getCmat(pdGIdGU_gv,10,10,nGener_gv,nGener_gv,cpu);    {dIg/dUg for generator ii}
//showmessage(comptostr(cpu));

    for iter := 1 to {20}niter do begin
      clearDmat(paa,nn);
      clearDvec(pbb,nn);
      clearIvec(pind,nn);
      clearDvec(pww,nn);
      if (nGener_gv > 0) then begin
        {construct the equations, first the equations for the generators}
        ppGener := ppGener_gv;
        for ii:=1 to nGener_gv do begin
          pGener := ppGener^;                   {pointer to this generator}
          inc(ppGener);
          {get latest generator voltage and current}
          GetVolt(pGener,pSolut_lf,1,nnod_gv,Ugen);  {get generator output voltage}
          GetVolt(pGener,pSolut_lf,2,nnod_gv,Igen);  {get generator current}
          {contribution from generators - - - - - - - - - - - - - - - -}
          {isolated generator, or a generator that cannot change the output voltage,
           mostly = parallel with a feeder}
          getCmat(pdGIdGU_gv,ii,ii,nGener_gv,nGener_gv,dGIdGUii);    {dIg/dUg for generator ii}
          getCmat(pdGUdGU_gv,ii,ii,nGener_gv,nGener_gv,cpu);         {singular, if dU/dU = 0 }
//showmessage(inttostr(ii)+'  '+floattostr(cabs(cpu)));
          iparal := 0;                                               {initialize}
          if (pGener^.int1_s = 0) and (Cabs(cpu) > 0.01) then begin  {not isolated generator}
            {get the index of eventual parallel generator}
            getIvec(pParal_gv,ii,nGener_gv,iparal);
            for jj:=1 to nGener_gv do begin
              {get (complex)derivatives }
              getCmat(pdGIdGU_gv,ii,jj,nGener_gv,nGener_gv,dGIdGU);
              getCmat(pdGUdGU_gv,ii,jj,nGener_gv,nGener_gv,dGUdGU);
//if (ii = jj) then showmessage(inttostr(ii)+'  '+floattostr(cabs(dgidgu))+'  '+floattostr(cabs(dgudgu)));
              {get (real) derivatives of real and imag. parts from these}
              {I and E refer to generator, V and J to load}
              dIrdEr := dGIdGU.x;
              dUrdEr := dGUdGU.x;
              dIidEr := dGIdGU.y;
              dUidEr := dGUdGU.y;
              dIrdEi :=-dGIdGU.y;
              dUrdEi :=-dGUdGU.y;
              dIidEi := dGIdGU.x;
              dUidEi := dGUdGU.x;
              {contribution to first generator equation = update coefficient matrix}
              apu := Ugen.x*dIrdEr+Igen.x*dUrdEr+Ugen.y*dIidEr+Igen.y*dUidEr;
              putDmat(paa,2*ii-1,2*jj-1,nn,nn,apu);
              apu := Ugen.x*dIrdEi+Igen.x*dUrdEi+Ugen.y*dIidEi+Igen.y*dUidEi;
              putDmat(paa,2*ii-1,2*jj  ,nn,nn,apu);
              {contribution to second generator equation}
              {must check eventual parallel generators}
              if (iparal <= 0) then begin    {no (earlier) parallel generators}
                apu := Ugen.x*dUrdEr+Ugen.y*dUidEr;
                putDmat(paa,2*ii,2*jj-1,nn,nn,apu);
                apu := Ugen.x*dUrdEi+Ugen.y*dUidEi;
                putDmat(paa,2*ii,2*jj  ,nn,nn,apu);
              end else begin
                {the equation is produced later, when parallel gener.}
              end;
            end;
            {the second generator equation, when there is a parallel generator
            Require that the (correction of the) internal voltage is near
            to the (correction of the) internal voltage
            of the parallel generator, while keeping the power at the desired value}
            if (iparal > 0) then begin
              if (iparal <= nGener_gv) then begin
                {coefficients were calculated for the first gener.equation}
                getDmat(paa,2*ii-1,2*ii-1    ,nn,nn, dpu);
                getDmat(paa,2*ii-1,2*ii      ,nn,nn, dpv);
                putDmat(paa,2*ii  ,2*ii-1    ,nn,nn, dpv);
                putDmat(paa,2*ii  ,2*iparal-1,nn,nn,-dpv);
                putDmat(paa,2*ii  ,2*ii      ,nn,nn,-dpu);
                putDmat(paa,2*ii  ,2*iparal  ,nn,nn, dpu);
              end else begin
                putDmat(paa,2*ii,2*ii,nn,nn,1.0);     {just some value}
              end;
            end;
          end else begin   {isolated generator}
//pGener^.GetLabel(lbl,ix,iy);
//showmessage(lbl+'  '+floattostr(cabs(cpu)));
            {if (dIg/dUg <> 0) try to get the power right, cannot do anything with the voltage}
            if (Cabs(dGIdGUii) > 0.01) then begin
              apu := Ugen.x*dGIdGUii.x+Ugen.y*dGIdGUii.y;
              apv := -Ugen.x*dGIdGUii.y+Ugen.y*dGIdGUii.x;
              putDmat(paa,2*ii-1,2*ii-1,nn,nn,apu);
              putDmat(paa,2*ii-1,2*ii  ,nn,nn,apv);
              apu := -Ugen.x*dGIdGUii.y+Ugen.y*dGIdGUii.x;
              apv := -Ugen.x*dGIdGUii.x-Ugen.y*dGIdGUii.y;
              putDmat(paa,2*ii,2*ii-1,nn,nn,apu);
              putDmat(paa,2*ii,2*ii  ,nn,nn,apv);
            end else begin  {truly isolated generator, ones to diagonal}
              putDmat(paa,2*ii-1,2*ii-1,nn,nn,1.0);
              putDmat(paa,2*ii,  2*ii  ,nn,nn,1.0);
            end;
          end;
          {contribution from loads, if there are any}
          if (nDrain_lf > 0) then begin
            for jj:=1 to nDrain_lf do begin
              {get (complex)derivatives }
              getCmat(pdGIdLI_lf,ii,jj,nGener_gv,nDrain_lf,dGIdLI);
              getCmat(pdGUdLI_lf,ii,jj,nGener_gv,nDrain_lf,dGUdLI);
              {get (real) derivatives of real and imag. parts from these}
              {I and E refer to generator, V and J to load}
              dIrdJr := dGIdLI.x;
              dUrdJr := dGUdLI.x;
              dIidJr := dGIdLI.y;
              dUidJr := dGUdLI.y;
              dIrdJi :=-dGIdLI.y;
              dUrdJi :=-dGUdLI.y;
              dIidJi := dGIdLI.x;
              dUidJi := dGUdLI.x;
              {contribution to first generator equation}
              apu := Ugen.x*dIrdJr+Igen.x*dUrdJr+Ugen.y*dIidJr+Igen.y*dUidJr;
              putDmat(paa,2*ii-1,2*nGener_gv+2*jj-1,nn,nn,apu);
              apu := Ugen.x*dIrdJi+Igen.x*dUrdJi+Ugen.y*dIidJi+Igen.y*dUidJi;
              putDmat(paa,2*ii-1,2*nGener_gv+2*jj  ,nn,nn,apu);
              {contribution to second generator equation}
              {only, if no parallel generators}
              if (iparal <= 0) then begin
                apu := Ugen.x*dUrdJr+Ugen.y*dUidJr;
                putDmat(paa,2*ii,2*nGener_gv+2*jj-1,nn,nn,apu);
                apu := Ugen.x*dUrdJi+Ugen.y*dUidJi;
                putDmat(paa,2*ii,2*nGener_gv+2*jj  ,nn,nn,apu);
              end;
            end;
          end;
          {calculate the right hand side = deviation from desired values}
          if (pGener^.int1_s = 0) or (Cabs(dGIdGUii) > 0.01)then begin  {not isolated}
//showmessage(floattostr(Cabs(dgidguii)));
            ngp := nGenP(pGener);                   {number of parallel generators}
            pGener^.GetPar(gGen_W,PGgoal);          {desired power}
            PGgoal := ngp*PGgoal;
            apu := (Ugen.x*Igen.x+Ugen.y*Igen.y);   {present power of one phase}
//showmessage(inttostr(iter)+'  '+floattostr((apu*3-pGgoal)*1e-6));
            if (PGgoal > 0) then begin
              putDvec(pbb,2*ii-1,nn,PGgoal/3.0-apu); {rhs = -error of one phase}
            end else begin
              putDvec(pbb,2*ii-1,nn,0.0);
              show_message('nonpositive desired power');
            end;
            pGener^.GetPar(gGen_Vn,apu);          {desired phase to phase voltage}
            U2goal := apu*apu/3.0;                {desired phase to neutral U squared}
            {normal, if no parallel generators}
            if (iparal <= 0) then begin
              apu := Ugen.x*Ugen.x+Ugen.y*Ugen.y;   {present U**2}
              putDvec(pbb,2*ii,nn,0.5*(U2goal-apu)); {rhs = -error}
            end else begin
            end;
          end else begin   {an isolated generator, zeroes to the rhs}
            putDvec(pbb,2*ii-1,nn,0.0);
            putDvec(pbb,2*ii  ,nn,0.0);
          end;
        end;
      end;
      {construct equations for the load, if there are any}
      if (nDrain_lf > 0) then begin
        ppLoad := ppDrain_lf;
        for ii:=1 to nDrain_lf do begin
          pLoad := ppLoad^;                   {pointer to this load}
          inc(ppLoad);
          {get latest generator voltage and current}
          GetVolt(pLoad,pSolut_lf,1,nnod_gv,Uload);     {get load voltage}
          GetVolt(pLoad,pSource_lf,2{1},nnod_gv,Iload); {get load current}
          {contribution from generators}
          if (nGener_gv > 0) then begin
            for jj:=1 to nGener_gv do begin
              {get (complex) derivatives }
              getCmat(pdLUdGU_lf,ii,jj,nDrain_lf,nGener_gv,dLUdGU);
              {get (real) derivatives of real and imag. parts from these}
              {I and E refer to generator, V and J to load}
              dVrdEr := dLUdGU.x;
              dVidEr := dLUdGU.y;
              dVrdEi :=-dLUdGU.y;
              dVidEi := dLUdGU.x;
              {contribution to first load equation}
              apu := Iload.x*dVrdEr+Iload.y*dVidEr;
              putDmat(paa,2*nGener_gv+2*ii-1,2*jj-1,nn,nn,apu);
              apu := Iload.x*dVrdEi+Iload.y*dVidEi;
              putDmat(paa,2*nGener_gv+2*ii-1,2*jj  ,nn,nn,apu);
              {contribution to second load equation}
              apu := Iload.x*dVidEr-Iload.y*dVrdEr;
              putDmat(paa,2*nGener_gv+2*ii,2*jj-1,nn,nn,apu);
              apu := Iload.x*dVidEi-Iload.y*dVrdEi;
              putDmat(paa,2*nGener_gv+2*ii,2*jj  ,nn,nn,apu);
            end;
          end;
          {contribution from loads}
          if (pLoad^.int1_s = 0) then begin  {not isolated load}
            for jj:=1 to nDrain_lf do begin
              {get (complex) derivatives }
              getCmat(pdLUdLI_lf,ii,jj,nDrain_lf,nDrain_lf,dLUdLI);
//if (ii = jj) then showmessage(inttostr(jj)+'  '+floattostr(cabs(dludli)));
              {get (real) derivatives of real and imag. parts from these}
              {I and E refer to generator, V and J to load}
              dVrdJr := dLUdLI.x;
              dVidJr := dLUdLI.y;
              dVrdJi :=-dLUdLI.y;
              dVidJi := dLUdLI.x;
              {contribution to first load equation}
              apu := Iload.x*dVrdJr+Iload.y*dVidJr;
              if (ii=jj) then apu := apu+Uload.x;          {contribution to itself}
              putDmat(paa,2*nGener_gv+2*ii-1,2*nGener_gv+2*jj-1,nn,nn,apu);
              apu := Iload.x*dVrdJi+Iload.y*dVidJi;
              if (ii = jj) then apu := apu+Uload.y;
              putDmat(paa,2*nGener_gv+2*ii-1,2*nGener_gv+2*jj  ,nn,nn,apu);
              {contribution to second load equation}
              apu := Iload.x*dVidJr-Iload.y*dVrdJr;
              if (ii=jj) then apu := apu+Uload.y;
              putDmat(paa,2*nGener_gv+2*ii,2*nGener_gv+2*jj-1,nn,nn,apu);
              apu := Iload.x*dVidJi-Iload.y*dVrdJi;
              if (ii=jj) then apu := apu-Uload.x;
              putDmat(paa,2*nGener_gv+2*ii,2*nGener_gv+2*jj  ,nn,nn,apu);
            end;
          end else begin        {isolated load, ones to diagonal}
            putDmat(paa,2*nGener_gv+2*ii-1,2*nGener_gv+2*ii-1,nn,nn,1);
            putDmat(paa,2*nGener_gv+2*ii,  2*nGener_gv+2*ii  ,nn,nn,1);
          end;
          {calculate the right hand side = deviation from desired values}
          {different parameters for (proper) load, and current drain}
          pLoad^.GetModName(ModNam);
          if (ModNam = gCurr) then begin            {drain}
            pLoad^.GetPar(gCurr_VA,VAgoal);         {desired power (VA)}
            pLoad^.GetPar(gCurr_cosf,cfgoal);       {desired cos(fii)}
          end else if (ModNam = gLoad) then begin   {load}
            apw := pSet_lf^.Load_mult; { if (apw <= 0) then apw := 1; }
//          pLoad^.GetPar(gLoad_VA,apu{VAgoal});
            pSet_lf^.GetLoadParam(pLoad,gLoad_VA,apu);
//          pLoad^.GetPar(gLoad_m,apv);
            pSet_lf^.GetLoadParam(pLoad,gLoad_m,apv);
            {if (apv > 0) then} VAgoal := apu*apv*apw {else VAgoal := apu*apw};
//          pLoad^.GetPar(gLoad_pf,cfgoal);
            pSet_lf^.GetLoadParam(pLoad,gLoad_pf,cfgoal);
          end else begin                            {unknown, should not be}
            VAgoal := 0;
            cfgoal := 0;
          end;
//        if (VAgoal > 0) and (cfgoal > 0) and (abs(cfgoal) <= 1) then begin
          if (abs(cfgoal) <= 1) then begin
            PLgoal := VAgoal*cfgoal/3.0;                {desired P of one phase}
            QLgoal := VAgoal*sqrt(1-cfgoal*cfgoal)/3.0; {desired Q of one phase}
            if (pLoad^.int1_s = 0) then begin  {not isolated}
              apu := (Uload.x*Iload.x+Uload.y*Iload.y);   {present real power}
              putDvec(pbb,2*nGener_gv+2*ii-1,nn,PLgoal-apu);
              apv := (Uload.y*Iload.x-Uload.x*Iload.y);   {present imag.power}
              putDvec(pbb,2*nGener_gv+2*ii  ,nn,QLgoal-apv);
            end else begin
              putDvec(pbb,2*nGener_gv+2*ii-1,nn,0);
              putDvec(pbb,2*nGener_gv+2*ii  ,nn,0);
            end;
          end else begin
            putDvec(pbb,2*nGener_gv+2*ii-1,nn,0.0);
            putDvec(pbb,2*nGener_gv+2*ii  ,nn,0.0);
            show_message('not reasonable load');
          end;
        end;
      end;
      {solve the corrections to generator voltage (phase) and load current}
//showDmat1(paa,nn,nn);
//showmessage(' ');
      {NOTE, solution overwrites right hand side}
//ShowDmat1(pbb,1,nn);
//showmessage(' ');
      DLUdecomp(paa ,pww, nn, pind, Dodd, Sngl, Tout);
      DLUbacksub(paa, nn, pind, pbb);
//ShowDmat1(pbb,1,nn);
      {change (correct) the source terms, solve the network again}
      if (nGener_gv > 0) then begin
        {generators}
        ppGener := ppGener_gv;
        for ii:=1 to nGener_gv do begin
          pGener := ppGener^;                       {pointer to generator}
          inc(ppGener);
          getVolt(pGener,pSource_lf,2,nnod_gv,cpu); {get old voltage}
          getDvec(pbb,2*ii-1,nn,dpu);               {get correction, real part}
          if (abs(dpu) < 1.e8) then begin
            cpv.x := dpu;
          end else begin
            if (dpu > 0) then cpv.x := 1.e8
            else cpv.x := -1.e8;
          end;
          getDvec(pbb,2*ii  ,nn,dpu);               {get correction, imago.part}
          if (abs(dpu) < 1.e8) then begin
            cpv.y := dpu;
          end else begin
            if (dpu > 0) then cpv.y := 1.e8
            else cpv.y := -1.e8;
          end;
          {cpv.y := dpu;}
          cpu := Cadd(cpu,cpv);                     {new voltage}
          putVolt(pGener,pSource_lf,2,nnod_gv,cpu); {put new voltage back}
        end;
      end;
      {loads, if any}
      if (nDrain_lf > 0) then begin
        ppLoad := ppDrain_lf;
        for ii:=1 to nDrain_lf do begin
          pLoad := ppLoad^;                         {pointer to load}
          inc(ppLoad);
          getVolt(pLoad,pSource_lf,2,nnod_gv,cpu);  {get old current}
          getDvec(pbb,2*nGener_gv+2*ii-1,nn,dpu);   {get correction, real part}
          if (abs(dpu) < 1.e8) then begin
            cpv.x := dpu;
          end else begin
            if (dpu > 0) then cpv.x := 1.e8
            else cpv.x := -1.e8;
          end;
{         cpv.x := dpu;}
          getDvec(pbb,2*nGener_gv+2*ii  ,nn,dpu);   {get correction, imag.part}
          if (abs(dpu) < 1.e8) then begin
            cpv.y := dpu;
          end else begin
            if (dpu > 0) then cpv.y := 1.e8
            else cpv.y := -1.e8;
          end;
{         cpv.y := dpu; }
          cpu := Cadd(cpu,cpv);                     {new current}
          putVolt(pLoad,pSource_lf,2,nnod_gv,cpu);  {put new current back}
        end;
      end;
      {copy source to a vector that will become the solution vector}
      copyCvec(pSource_lf,pSolut_lf,nnod_gv);
      {solve again = backsubstitute, with these source terms}
      CLUbacksub(pAdmat,nnod_gv,pIndex,pSolut_lf);{get "solution" }
      {check convergence. Calculate total correction}
      CorrSum := 0;
      line := '';
      for ii:=1 to nn do begin
         getDvec(pbb,ii,nn,dpu);
         line := line+'   '+floattostr(dpu);
         CorrSum := CorrSum+abs(dpu);
      end;
// showmessage(inttostr(iter)+'  '+floattostr(CorrSum*1e-6)+'  '+floattostr(nn*ccrit));
//showmessage(inttostr(iter)+'  '+line);
      if (CorrSum/nn <= ccrit) {or (iter >= 25)} then begin
        converged := true;
        break;
      end else begin
        converged := false;
      end;
    end;                                          {end of the iterations}
    {announce eventual nonconvergence}
    if (not converged) then begin
      show_message('load flow did not converge');
      CopyCvec(pNomSol_gv,pRhs,nnod_gv);          {take the initial guess}
    end else begin
      {copy results}
      CopyCvec(pSolut_lf,pRhs,nnod_gv);
    end;
    {save the result as such, needed for superposition, in SolveNet of elplex}
    pSet^.List_Probl := probl;                           {save the problem type..}
    pSet^.SaveSolution(pSet^.Volx_List,pSet^.Voly_List); {..and the solution}
    {if did not converge, show the initial solution}
    SaveAndDisplayResults(pSet,nnod_gv,freq_gv);
    {copy the results for later use, in short circuits}
    CopyPosSeqResults(pSet);
    pSet^.SavePrefaultVoltages('pref');        {save for buses, voltmeters, shorts}
    {finally, release the reserved memory}
    freemem(paa,nn*nn*SizeOf(double));
    freemem(pbb,nn*SizeOf(double));
    freemem(pww,nn*SizeOf(double));
    freemem(pind,nn*SizeOf(integer));
  end;

end;

procedure unconnected(pSet: TpCompSet);
{find unconnected loads and generators: current (nearly) zero
 NOTE: network must have been solved}
var
  apu,apv : single;  {temporary}
  curr    : complex; {current to load}
  errc    : integer; {error flag}
//ii : integer;
ix,iy   : integer;
  jcomp   : integer; {counts components}
lbl     : string;
  ModNam  : string;  {name (type) of the component}
  ngp     : single;  {number of parallel generators}
  pComp   : TpComp;  {pointer to a component}
  pow     : single;  {power}
  volt    : single;  {phase to phase voltage}
  smalcur : single;  {a small current}
begin
  {scan all components, find loads}
  errc := 0;
  pComp := nil;
  jcomp := 0;
//  ii := 0;
  while (errc = 0) do begin                {scan all components}
    pSet^.NextComp(pComp,errc,jcomp,true); {get pointer to next component}
    if (errc = 0) then begin               {component found}
      pComp^.GetModName(ModNam);           {get the name (type) of the component}
pComp^.GetLabel(lbl,ix,iy);
      if (ModNam = gLoad) then begin       {is it a drain (load)}
        pComp^.GetVar(gLoad_Ir,curr.x);    {real part of current}
        pComp^.GetVar(gLoad_Ii,curr.y);    {imag. part of current}
//inc(ii);
//showmessage(inttostr(ii)+'  '+lbl+'  '+floattostr(cabs(curr)));
        if (Cabs(curr) < 0.01) then pComp^.int1_s := 1  {set flag}
        else pComp^.int1_s := 0;
      end else if (ModNam = gGen) then begin  {generator}
        pComp^.GetVar(gGen_Ir,curr.x);        {real part of current}
        pComp^.GetVar(gGen_Ii,curr.y);        {imag. part of current}
        {the current of a generator might not be very small due to transformers, for example}
        ngp := nGenP(pComp);                  {number of parallel generators}
        pComp^.GetPar(gGen_W,pow);            {desired power}
        pow := pow*ngp;
        pComp^.GetPar(gGen_Vn,volt);          {nominal phase to phase voltage}
        if (pow > 0) and (volt > 0) then begin
          apu := 0.07*pow/(volt*sqrt(3));
          apv := volt/80;                     {7.3.07 for small generators LFtest.sld}
          if (apv < apu) then apu := apv;
          if (apu > 1) then smalcur := apu else smalcur := 1;
        end else begin
          smalcur := 1;
        end;
//showmessage(lbl+'  '+floattostr(cabs(curr))+'  '+floattostr(smalcur));
        if (Cabs(curr) < smalcur) then pComp^.int1_s := 1  {set flag}
        else pComp^.int1_s := 0;
      end;
    end;
  end;
end;

procedure LoadFlowCalculation(ipSet : longint; probl: Solve; var success : boolean);  {solve defined in elplex.pas}
{solve the load flow problem for the net
 ipSet   pointer to the set of components etc, as integer
 success true, if successful calculation}
var
  isDecomposed : boolean;   {true, if the admittance matrix has been decomposed}
  pSet         : TpCompSet; {pointer to the "common"}
{ success      : boolean;   {flags that load flow can be started}
begin
  pSet := TpCompSet(ipSet);
  {solve net in order to find unconnected loads}
  if not(probl = TapLoad) then begin    {already done in TapChanger.pas}
    Imagoo.AdmatOper(false);
    Unconnected(pSet);
  end;
  {create obj.for load flow calculations}
  LoadFlower := TLoadFlow.Create(ipSet,success);
  {continue, if there are generator(s) etc, i.e. if successful}
  if (success) then begin
    pSet^.GetMemForCurrTrAndRelays;    {meory for curr.tr and relays (if needed)}
    {equations for the network}
    LoadFlower.ReserveConstructAndDecompose({SolvingLoadFlow}probl,isDecomposed);
    {continue only, if managed to decompose the equation matrix}
    if (isDecomposed) then begin
      {solve the network using the given generator voltages}
      LoadFlower.FindNominalSolution;                        {in GenVolt.pas}
      {calculate derivatives, needed in the solution process}
      if (LoadFlower.nGener_gv > 0) then begin
        LoadFlower.CalculateGeneratorVoltageDerivative;
        LoadFlower.CalculateOtherGeneratorDerivatives;
      end;
      LoadFlower.CalculateLoadDerivatives;
      if (LoadFlower.nGener_gv > 0) then begin
        {find eventual parallel generators (one, at most, for each}
        LoadFlower.FindParallelGenerator;
      end;
      {solve the load flow problem}
      LoadFlower.SolveLoadFlow(probl);
    end else begin
      show_message('error in load flow: coeff. matrix ?');
    end;
    {free the reserved memory from matrices}
    FreeMat(LoadFlower.nnod_gv);
    pSet^.FreeMemForCurrTrAndRelays;    {meory for curr.tr and relays (if needed)}
  end else begin
{    show_message('No generator or no load, thus no calculation');}
  end;
  {finally, free the object for load flow calculations}
  LoadFlower.free;                {free the object for l-f calculations}

end;

end.
